import configparser
from pathlib import Path
from unittest import TestCase

from bson import DBRef, ObjectId
from edman import DB, Config
from pymongo import MongoClient
from pymongo import errors as py_errors

from edman_web.search_manager import SearchManager


class TestSearchManager(TestCase):
    db_server_connect = False
    test_ini: dict = {}
    client = None

    @classmethod
    def setUpClass(cls):
        # 設定読み込み
        settings = configparser.ConfigParser()
        settings.read(Path.cwd() / 'ini' / 'test_db.ini')
        cls.test_ini = dict(settings.items('DB'))
        cls.test_ini['port'] = int(cls.test_ini['port'])

        # DB作成のため、pymongoから接続
        cls.client = MongoClient(cls.test_ini['host'], cls.test_ini['port'])

        # 接続確認
        try:
            cls.client.admin.command('hello')
            cls.db_server_connect = True
            print('Use DB.')
        except py_errors.ConnectionFailure:
            print('Do not use DB.')

        if cls.db_server_connect:
            # adminで認証
            cls.client = MongoClient(
                username=cls.test_ini['admin_user'],
                password=cls.test_ini['admin_password'])
            # DB作成
            cls.client[cls.test_ini['db']].command(
                "createUser",
                cls.test_ini['user'],
                pwd=cls.test_ini['password'],
                roles=[
                    {
                        'role': 'dbOwner',
                        'db': cls.test_ini['db'],
                    },
                ],
            )
            # edmanのDB接続オブジェクト作成
            con = {
                'host': cls.test_ini['host'],
                'port': cls.test_ini['port'],
                'user': cls.test_ini['user'],
                'password': cls.test_ini['password'],
                'database': cls.test_ini['db'],
                'options': [f"authSource={cls.test_ini['db']}"]
            }
            db = DB(con)
            cls.testdb = db.get_db
            cls.search_manager = SearchManager(db)
        # else:
        #     cls.search = Search()

    @classmethod
    def tearDownClass(cls):
        if cls.db_server_connect:
            # cls.clientはpymongo経由でDB削除
            # cls.testdb.dbはedman側の接続オブジェクト経由でユーザ(自分自身)の削除
            cls.client.drop_database(cls.test_ini['db'])
            cls.testdb.command("dropUser", cls.test_ini['user'])

    def setUp(self):
        self.config = Config()
        self.parent = self.config.parent
        self.child = self.config.child
        self.date = self.config.date
        self.file = self.config.file

    def tearDown(self):
        if self.db_server_connect:
            # システムログ以外のコレクションを削除
            collections_all = self.testdb.list_collection_names()
            log_coll = 'system.profile'
            if log_coll in collections_all:
                collections_all.remove(log_coll)
            for collection in collections_all:
                self.testdb.drop_collection(collection)

    def test_get_documents(self):

        # テストデータ入力
        if not self.db_server_connect:
            return
        # docをDBに入れる
        parent_id = ObjectId()
        doc_id = ObjectId()
        child_id = ObjectId()
        parent_col = 'parent_col'
        doc_col = 'doc_col'
        child_col = 'child_col'
        insert_docs = [
            {
                'col': parent_col,
                'doc': {
                    '_id': parent_id,
                    'name': 'parent',
                    Config.child: [DBRef(doc_col, doc_id)]
                },
            },
            {
                'col': doc_col,
                'doc': {
                    '_id': doc_id,
                    'name': 'doc',
                    Config.parent: DBRef(parent_col, parent_id),
                    Config.child: [DBRef(child_col, child_id)]
                }
            },
            {
                'col': child_col,
                'doc': {
                    '_id': child_id,
                    'name': 'child',
                    Config.parent: DBRef(doc_col, doc_id),
                }
            }]
        result = {}
        for i in insert_docs:
            insert_result = self.testdb[i['col']].insert_one(i['doc'])
            result.update({i['col']: insert_result.inserted_id})

        all_docs = self.search_manager.get_documents(2, doc_col, doc_id,
                                                     parent_depth=0,
                                                     child_depth=0)
        expected = {
            parent_col: {
                'name': 'parent',
                doc_col: [{
                    'name': 'doc',
                    child_col: [{'name': 'child'}]
                }]
            }
        }
        self.assertDictEqual(expected, all_docs)



import base64
import mimetypes
import os
from io import BytesIO
from typing import Any, List, Optional, Tuple, Union

import gridfs
from bson import ObjectId
from edman import Config, File
from edman.exceptions import EdmanDbProcessError, EdmanInternalError
from edman.utils import Utils
from gridfs.errors import GridFSError
from PIL import Image as PILImage
from werkzeug.datastructures import FileStorage


class FileManager(File):
    def __init__(self, db=None):
        super().__init__(db)

    def web_upload(self, collection: str, oid: Union[str, ObjectId],
                   up_file: FileStorage, compress=False) -> None:
        """
        ファイルアップロード処理

        :param str collection:
        :param str or ObjectId oid:
        :param FileStorage up_file:
        :param bool compress:
        :return:
        """
        oid = Utils.conv_objectid(oid)

        # ドキュメント存在確認&対象ドキュメント取得
        if (doc := self.db[collection].find_one({'_id': oid})) is None:
            raise EdmanDbProcessError('対象のドキュメントが存在しません')

        try:
            # gridfsにファイルを入れる
            inserted_file_oids = self.web_grid_in(up_file, compress)
        except EdmanDbProcessError as e:
            raise e
        else:  # ドキュメントの更新
            try:
                new_doc = self.file_list_attachment(doc, inserted_file_oids)
                replace_result = self.db[collection].replace_one({'_id': oid},
                                                                 new_doc)
                if replace_result.modified_count != 1:
                    # ドキュメントが更新されていない場合はgridfsからデータを削除する
                    self.fs_delete(inserted_file_oids)
                    raise EdmanDbProcessError('ドキュメントの更新ができませんでした.')
            except Exception as e:
                # 途中で例外が起きた場合、gridfsからデータを削除する
                self.fs_delete(inserted_file_oids)
                raise EdmanDbProcessError(str(e))

    def web_grid_in(self, file: FileStorage, compress: bool) -> list[Any]:
        """
        Gridfsへデータをアップロード

        :param FileStorage file:
        :param bool compress:
        :return: inserted
        :rtype: list
        """
        inserted = []
        try:
            f = file.stream.read()
            metadata = {'filename': file.filename}
        except OSError:
            raise EdmanDbProcessError('DBにファイルをアップロード出来ませんでした')
        except Exception:
            raise
        try:
            inserted.append(self.fs.put(f, **metadata))
        except GridFSError as e:
            raise EdmanDbProcessError(e)
        return inserted

    def file_download(self, oid: Union[ObjectId, str]
                      ) -> tuple[gridfs.grid_file.GridOut, str, Optional[str]]:
        """
        GridFsからファイルをダウンロードする

        :param str or ObjectId oid:
        :rtype: tuple
        :return:
        """
        if not isinstance(oid, ObjectId):
            if ObjectId.is_valid(oid):
                oid = ObjectId(oid)
            else:
                raise ValueError('ObjectIdに合致しません')

        # ファイル情報を取得
        try:
            content = self.fs.get(oid)
        except gridfs.errors.NoFile:
            raise ValueError('ファイルが存在しません')
        except gridfs.errors.GridFSError:
            raise
        file_name = content.filename
        mimetype = mimetypes.guess_type(file_name)[0]
        return content, file_name, mimetype

    def file_delete(self, collection: str, oid: Union[str, ObjectId],
                    delete_list: List[str]):
        """
        edmanからファイルを削除する

        :param str collection:
        :param str or ObjectId oid:
        :param list delete_list:
        :return:
        """
        oid = Utils.conv_objectid(oid)

        # ドキュメント存在確認&対象ドキュメント取得
        if (doc := self.db[collection].find_one({'_id': oid})) is None:
            raise EdmanDbProcessError('対象のドキュメントが存在しません')
        if not delete_list:
            raise EdmanInternalError('削除対象リストが存在しません')

        delete_items = tuple((map(lambda x: ObjectId(x), delete_list)))

        # ファイルリファレンスから指定のoidを削除する
        diff = list(set(doc[Config.file]) - set(delete_items))
        try:
            replace_doc = self.file_list_replace(doc, diff)
            replace_result = self.db[collection].replace_one(
                {'_id': oid}, replace_doc)
            if replace_result.modified_count != 1:
                # ドキュメントが更新されていない場合
                raise EdmanDbProcessError(
                    f'ファイルリファレンスを削除できませんでした.{diff} ファイルは削除されません')
        except Exception:
            raise
        else:
            # ファイルリファレンスの削除が成功した場合のみ、gridfsからデータを削除する
            try:
                self.fs_delete(list(delete_items))
            except Exception:
                raise

    @staticmethod
    def extract_thumb_list(files: list, thumbnail_suffix: list
                           ) -> List[Tuple[ObjectId, str]]:
        """
        サムネ作成対象のリストを作成する

        :param list files:
        :param list thumbnail_suffix:
        :return:
        :rtype: list
        """
        j = [os.path.splitext(i[1])[1][1:] for i in files]
        return [(files[idx][0], ext) for idx, ext in enumerate(j) if
                ext in thumbnail_suffix]

    @staticmethod
    def generate_thumbnail(content: bytes, ext: str,
                           thumbnail_size: tuple[int, int],
                           file_decode='utf-8') -> str:
        """
        サムネイル画像をbase64で作成

        :param bytes content:
        :param str ext:
        :param tuple thumbnail_size:
        :param str file_decode: default 'utf-8'
        :return:
        :rtype: str
        """
        try:
            img = PILImage.open(BytesIO(content))
            img.thumbnail(size=thumbnail_size, resample=PILImage.LANCZOS)
            thumbnail = BytesIO()
            # jpgという拡張子は利用できないので変換する
            img.save(thumbnail, 'jpeg' if ext == 'jpg' else ext)
        except (IOError, KeyError) as e:
            raise EdmanInternalError(f'サムネイルが生成できませんでした {e}')
        try:
            outputfile = base64.b64encode(thumbnail.getvalue()).decode(
                file_decode)
        except Exception:
            raise
        return outputfile

    def get_thumbnails_procedure(self, files: list, thumbnail_suffix: list,
                                 thumbnail_size=(100, 100)) -> dict:
        """
        データをDBから出してサムネイルを取得するラッパー

        :param list files:
        :param list thumbnail_suffix:
        :param tuple thumbnail_size: default (100, 100)
        :return:
        :rtype: dict
        """
        thumbnails = {}
        for oid, ext in self.extract_thumb_list(files, thumbnail_suffix):
            # contentを取得
            try:
                content, file_name, mimetype = self.file_download(oid)
                content_data = content.read()
            except ValueError:
                raise
            try:
                # サムネイルを作成
                image_data = self.generate_thumbnail(content_data, ext,
                                                     thumbnail_size)
            except Exception:
                raise
            else:
                thumbnails.update({oid: {'data': image_data, 'suffix': ext}})

        return thumbnails



from typing import Union

from bson import ObjectId
from edman import Search
from edman.json_manager import GetJsonStructure


class SearchManager(Search):
    def __init__(self, db=None):
        super().__init__(db)

    def get_documents(self, dl_select: int, collection_name: str,
                      oid: Union[ObjectId, str], parent_depth: int,
                      child_depth: int, exclusion=None) -> dict:
        """
        指定したドキュメントをDBから取得する
        :param int dl_select:
        :param str collection_name:
        :param ObjectId or str oid:
        :param int parent_depth:
        :param int child_depth:
        :param List or None exclusion:
        :return: result
        :rtype: dict
        """
        if not isinstance(oid, ObjectId):
            if ObjectId.is_valid(oid):
                oid = ObjectId(oid)
            else:
                raise ValueError('ObjectIdに合致しません')

        # 階層指定
        if dl_select == GetJsonStructure.manual_select.value:
            result = self.find(collection_name, {'_id': ObjectId(oid)},
                               parent_depth=parent_depth,
                               child_depth=child_depth, exclusion=exclusion)

        # 自分が所属するツリー全て
        elif dl_select == GetJsonStructure.all_doc.value:
            doc = self.db.doc(collection_name, ObjectId(oid), query=None,
                              reference_delete=False)

            # 自分の所属するツリーのroot情報を取得
            if (
                    root_dbref := self.db.get_root_dbref(
                        doc)) is None:  # 自分自身がrootの場合
                root_collection = collection_name
                root_doc = doc
            else:
                root_collection = root_dbref.collection
                root_doc = self.connected_db[root_collection].find_one({
                    '_id': root_dbref.id
                })

            # 自分と子要素をマージ
            root_doc.update(
                self.db.get_child_all({root_collection: root_doc}))

            # リファレンス削除や日時変換
            result = self.process_data_derived_from_mongodb(
                {root_collection: root_doc}, exclusion=exclusion)

        else:
            # 単一のドキュメント
            result = self.find(collection_name, {'_id': ObjectId(oid)},
                               parent_depth=0, child_depth=0,
                               exclusion=exclusion)
        return result
